__package__    = "turkishMMap/__init__.py"
__version__    = "0.1"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/turkishMMap"
__date__       = "$Date: 2009/05/09 17:04:00 $"
__copyright__  = "Copyright (c) 2009 Aaron Straup Cope. BSD license : http://www.modestmaps.com/license.txt"

import PIL.Image
import PIL.ImageDraw

import ModestMaps
from reportlab.pdfgen import canvas

from reportlab.graphics.shapes import *
from reportlab.lib import colors
 
from modestMMarkers import polylines
from modestMMarkers import primitives
from modestMMarkers import text
from modestMMarkers import maps
from modestMMarkers import utils

import math

class turkishMMap :

    def __init__ (self, paper_width, paper_height, paper_margin=0.5, paper_dpi=144) :

        self.paper_height = paper_height
        self.paper_width = paper_width
        self.paper_margin = paper_margin
        self.paper_dpi = paper_dpi

        self.width = int(self.paper_width * self.paper_dpi)
        self.height = int(self.paper_height * self.paper_dpi)  
        self.margin = int(paper_margin * self.paper_dpi)

        self.orientation = 'portrait'

        if self.width > self.height:
            self.orientation = 'landscape'
            
        self.min_dim = min(self.width, self.height)

        self.map_width = self.min_dim
        self.map_height = self.min_dim

        self.map_offset_x = 0
        self.map_offset_y = self.height - self.map_height

        self.draw_fold_lines = True
        self.draw_howto_graphic = True

        self.draw_navi_graphic = False
        self.navi_graphic_bbox = None
        self.navi_graphic_context = None        

        self.mm_provider_name = None        
        self.mm_provider = None

        self.pages = []

        self.__mmobjs = []
        self.__markers = []
        
        # things that may be 'reset'

        self.reset()
        
    # ##########################################################
            
    def load_provider (self, label) :
        self.mm_provider_name = label
        self.mm_provider = ModestMaps.builtinProviders[label]()

    # ##########################################################

    def setup (self, bbox=None) :
        
        # can has markers?
        
        count_markers = len(self.markers)

        # check to see if we have a bbox and if
        # not try to use the markers to work it
        # out
        
        if not bbox:

            if count_markers == 0:
                return False

            allpoints = []
            
            for m in self.markers :
                allpoints.append({'lat':m['latitude'], 'lon':m['longitude']})
                
            bbox = calculate_bbox_for_points(allpoints)
            bbox = expand_bbox(bbox, .1)   
            
        # check to see if we have any markers and
        # adjust the height of the map accordingly

        if count_markers :

            if count_markers > 9 :
                self.legend_rows = int(math.ceil(float(count_markers) / float(self.legend_cols)))
            
            # note the +/- self.margin hoop jumping
            # when setting map/legend height...

            self.legend_height = int(self.min_dim * .25) + int(self.margin * 1.5)
            self.legend_width = self.map_width - (self.margin * 2)

            self.legend_item_height = int((self.legend_height - int(self.margin * 1.5)) / self.legend_rows)
            self.legend_item_width = int(self.legend_width / self.legend_cols)
            
            self.map_height -= self.legend_height

            self.legend_x_offset = self.margin
            self.legend_y_offset = self.height - (self.legend_height - int(self.margin / 2))

        self.bbox = bbox
                
    # ##########################################################

    def draw (self, bbox=None, out=None) :

        self.setup(bbox)
        
        self.draw_map(bbox)
        self.draw_boxes()
        self.draw_markers()
        self.draw_canvas()
        self.draw_legend()
        self.draw_navi_map()
        self.draw_page_title()
        self.draw_folds()

        self.pages.append(self.canvas)
        
        if out :
            self.save(out)

    # ##########################################################

    def draw_canvas (self) :

        w = self.width
        h = self.height
        
        self.canvas = PIL.Image.new('RGBA', (w, h), 'white')
        self.canvas.paste(self.mm_img, (self.map_offset_x, self.map_offset_y))
        
    # ##########################################################
    
    def draw_map (self, bbox) :
        
        sw = ModestMaps.Geo.Location(bbox[0], bbox[1])
        ne = ModestMaps.Geo.Location(bbox[2], bbox[3])
        dims = ModestMaps.Core.Point(self.map_width, self.map_height)

        mm_obj = ModestMaps.mapByExtent(self.mm_provider, sw, ne, dims)
        mm_img = mm_obj.draw()
        mm_img = mm_img.convert('RGBA')

        self.mm_obj = mm_obj        
        self.mm_img = mm_img

    # ##########################################################

    def draw_navi_map (self) :

        if not self.draw_navi_graphic :
            return

        if self.orientation == 'portrait' :

            offset = self.margin
            
            h = self.height - self.min_dim - int(offset / 2)
            w = self.min_dim
            
            x_off = 0
            y_off = 0
            
        else :
            pass
        
        bbox = self.navi_graphic_bbox
        ctx = self.navi_graphic_context

        mm_img = maps.mk_navi_map(bbox, ctx, w, h, self.mm_provider_name)
        self.canvas.paste(mm_img, (x_off, y_off))
            
    # ##########################################################
    
    def draw_folds (self) :

        return
    
        if not self.draw_fold_lines:
            return

        if self.orientation == 'portrait' :
            x1 = 0 
            y1 = self.height - self.min_dim
            x2 = self.width
            y2 = self.height

        else :
            x1 = 0
            y1 = 0
            x2 = self.min_dim
            y2 = self.height

        #
        
        
        dr = PIL.ImageDraw.Draw(self.canvas)

        # draw the cut line

        if self.orientation == 'portrait' :
            dr.line((x1, y1, x2, y1), fill=128)
        else :
            dr.line((x2, y1, x2, y2), fill=128)
            
        # draw the vertical crease down the middle
        # ensure that the map opens with the 'wings'
        # on the side and the legend on the bottom

        dr.line((int(x2 / 2), y1, int(x2 / 2), y2), fill=128)
            
        # draw top-left to bottom-right

        dr.line((x1, y1, x2, y2), fill=128)
            
        # draw bottom-left to top-right

        dr.line((x1, y2, x2, y1), fill=128)
        
    # ##########################################################

    def draw_boxes (self) :

        if not len(self.boxes) :
            return 

        poly = polylines.polyline(self.mm_obj)
        self.mm_img = poly.draw_polylines(self.mm_img, self.boxes, colour=(.005, 0, 1), no_fill=1)
        
    # ##########################################################
    
    def draw_markers (self) :

        if not len(self.markers) :
            return

        # PDF markers for a future release
        # self.__markers.append(self.markers)
        
        from modestMMarkers import modestMMarkers

        bg_c = (255, 255, 255)
        str_c = (0, 0, 0)
        txt_c = (0, 0, 0)
            
        mm = modestMMarkers(self.mm_obj)
        (self.mm_img, self.legend) = mm.plot_markers(self.mm_img, self.markers, background_c=bg_c, stroke_c=str_c, text_c=txt_c)
        
    # ##########################################################
    
    def draw_legend (self) :

        if not self.legend :
            return
        
        width = self.legend_item_width
        height = self.legend_item_height
        
        col_offset = 0
        row_offset = 0

        bg_c = (.5, 0, 0)
        str_c = (1, 1, 1)
        txt_c = (1, 1, 1)

        for data in self.legend :

            txt = data['label']
            
            if data.has_key('description') :
                txt = txt +  "\n" + data['description']
                
            # draw cairo text block with padding
            # circle/marker into the margin
            # convert to PIL

            font_face = "Helvetica Bold"
            font_size = 12

            c = text.draw_text_block(txt, width, height,
                                     padding_left=40, padding_right=20, padding_bottom=10,
                                     font_size=font_size, font_face=font_face)

            args = {'x':17, 'y':12, 'radius':10, 'string':data['id']}

            primitives.draw_circle_with_string(c, args)

            txt_img = utils.cairo2pil(c)
            sz = txt_img.size

            # paste!

            x_off = self.legend_x_offset + (width * col_offset)
            y_off = self.legend_y_offset + (height * row_offset)
            
            self.canvas.paste(txt_img, (x_off, y_off))

            row_offset += 1

            if col_offset == self.legend_cols :
                col_offset = 0
                
            if row_offset == self.legend_rows :
                col_offset += 1
                row_offset = 0
            
    # ##########################################################

    def draw_page_title (self) :

        if not self.page_title :
            return
        
        height = int(self.margin / 2)
        width = int(self.min_dim / 2)

        p_left = self.margin
        p_top = int(self.margin / 5)
        
        c = text.draw_text_block_to_PIL(self.page_title, width, height, font_size=20, padding_left=p_left, padding_top=p_top, ensure_width=True)

        y = self.map_offset_y
        self.canvas.paste(c, (0, y))
        
    # ##########################################################
    
    def save (self, out) :

        # reportlab docs say out can be any IO
        # thingy or a path
        
        # note the part where we (re) flip x, y
        # and height, width; see notes in setup()

        width = self.paper_width
        height = self.paper_height

        width *= self.paper_dpi
        height *= self.paper_dpi

        pdf = canvas.Canvas(out, pagesize=(width, height))
        i = 0

        for pg in self.pages :

            pdf.drawInlineImage(pg, 0, 0)

            # PDF markers for a future release
        
            """
            if i >= 1 :
                j = i - 1                
                mm_obj = self.__mmobjs[i]           
                self.draw_pdf_markers(pdf, mm_obj, self.__markers[j])
            """
            
            pdf.showPage()
            i += 1
            
        pdf.save()

    # ##########################################################

    def draw_pdf_markers (self, pdf, mm_obj,markers) :
        
        for m in markers :
            self.draw_pdf_marker(pdf, mm_obj, m)

    # ##########################################################
    
    def draw_pdf_marker (self, pdf, mm_obj, marker, i=1) :
        
        loc = ModestMaps.Geo.Location(marker['latitude'], marker['longitude'])
        pt = mm_obj.locationPoint(loc)  

        x = int(pt.x)
        y = int(pt.y)

        w = self.map_height
        h = self.map_height + self.legend_height
        
        y = h - y
        
        pdf.setFont("Helvetica", 12)
        pdf.setLineWidth(3)
        
        anchor = 25
        radius = 15
        
        pdf.setStrokeColor(colors.black)         
        pdf.setFillColor(colors.white)

        pdf.line(x, y, x, (y + anchor))
        pdf.circle(x, (y + anchor + radius), 15, stroke=1, fill=1)
        
        pdf.setFillColor(colors.black)
        pdf.drawString(x, (y + anchor + radius), unicode(i))
        
    # ##########################################################

    def add_boxes (self, boxes) :
        for b in boxes:
            self.boxes.append(b)
            
    # ##########################################################
    
    def add_markers (self, lotsofdata) :

        for data in lotsofdata :
            self.add_marker(data)

    # ##########################################################

    def add_marker (self, data) :

        required = ('latitude', 'longitude')
        
        for item in required :

            if not data.has_key(item) :
                # error
                return False

        self.markers.append(data)
    
    # ##########################################################    

    def reset (self) :
        
        self.bbox = None
        self.canvas = None
        self.page_title = None
        
        self.map_width = self.min_dim
        self.map_height = self.min_dim
        
        self.mm_obj = None
        self.mmarker_obj = None
        self.mm_img = None

        self.boxes = []
        self.markers = []
        self.legend = None
        
        self.legend_height = 0
        self.legend_cols = 3
        self.legend_rows = 3

        self.legend_x_offset = 0
        self.legend_y_offset = 0
        
        self.legend_item_width = 0
        self.legend_item_height = 0
